@@ -7,10 +7,18 @@ use warnings;
 use feature 'signatures';
 
 use Carp;
+use List::Util qw( any );
+use POSIX qw( ceil );
 use X11::XCB ':all';
 use X11::korgwm::Common;
 use Glib::Object::Introspection;
-use Gtk3 -init;
+use Gtk3;
+
+unless ($X11::korgwm::gtk_init) {
+    Gtk3::disable_setlocale();
+    Gtk3::init();
+    $X11::korgwm::gtk_init = 1;
+}
 
 my $display;
 my $win_expose;
@@ -18,9 +26,10 @@ my $font;
 my ($color_fg, $color_bg, $color_expose);
 my ($color_gdk_fg, $color_gdk_bg, $color_gdk_expose);
 
-sub _create_thumbnail($scale, $pixbuf, $title, $id, $cb) {
+sub _create_thumbnail($scale, $win, $title, $id, $cb) {
     my $vbox = Gtk3::Box->new(vertical => 0);
     my $vbox_inner = Gtk3::Box->new(vertical => 0);
+    my $pixbuf = $win->_get_pixbuf();
 
     # Normalize size
     my ($w, $h) = ($pixbuf->get_width(), $pixbuf->get_height());
@@ -77,8 +86,33 @@ sub expose {
     # Drop any previous window
     return if $win_expose;
 
-    # Update pixbufs for all visible windows
-    $_->_update_pixbuf() for map { $_->current_tag()->windows() } values %screens;
+    # We should return earlier unless windows exist
+    my $nwindows = keys %{ $windows };
+    return unless $nwindows;
+
+    # If there is only one window we just want to focus it.
+    # Sorry for this ugly code. The logic is mostly copied from focus_prev() of Executor
+    if ($nwindows == 1) {{
+        my $win = (values %{ $windows })[0];
+        return carp "Unable to find single existing window to focus" unless $win;
+
+        my @tags = $win->tags();
+        my $tag = shift @tags // ($win->{always_on} && $win->{always_on}->current_tag());
+
+        # "Window $win is visible on multiple tags, do not know how to focus_prev() to it" so return to main routine
+        last if @tags;
+
+        return carp "Window $win has no tags and is not always_on" unless $tag;
+
+        # Switch to proper tag unless it is already active
+        unless (any { $tag == ($_->current_tag() // 0) } @screens) {
+            $tag->{screen}->{focus} = $win;
+            $tag->{screen}->tag_set_active($tag->{idx});
+            $tag->{screen}->refresh();
+        }
+
+        $win->warp_pointer();
+    }}
 
     # Select current screen
     my $screen_curr = $focus->{screen};
@@ -106,11 +140,15 @@ sub expose {
 
     # Estimate sizes
     # XXX it is incorrect as one window could belong to several tags, dont know how to represent it so left it for now
-    my $windows = keys %{ $windows };
-    return unless $windows;
-    my $rownum = _get_rownum($windows, @{ $screen_curr }{qw( w h )});
+    my $rownum = _get_rownum($nwindows, @{ $screen_curr }{qw( w h )});
     my $scale = 0.9 * ($screen_curr->{h} - $rownum * 2 * $cfg->{expose_spacing}) / $rownum;
-    my $id_len = 1 + int($windows / 10);
+
+    # This math formulae describes proper number of characters in window ID: 9 => 1, 10 => 2, 89 => 2, 90 => 3
+    my $lgnwindows = ceil(log($nwindows)/log(10));
+    my $id_len = $nwindows <= 9 ? 1 :
+        $nwindows <= (10 ** $lgnwindows - 10 ** ($lgnwindows - 1)) ?
+        $lgnwindows : $lgnwindows + 1;
+
     my %callbacks;
     my $shortcut_str = "";
     my $shortcut = Gtk3::Label->new();
@@ -139,11 +177,8 @@ sub expose {
                 my $id_str = "[$id]"; # should be string to avoid {0} === {"0"}
                 $callbacks{$id_str} = $cb if $cfg->{expose_show_id};
 
-                # If we never created a pixbuf for it
-                $win->_update_pixbuf() unless $win->{pixbuf};
-
                 # Create thumbnail
-                my $ebox = _create_thumbnail($scale, $win->{pixbuf}, $win->title(), $id_str, sub ($obj, $e) {
+                my $ebox = _create_thumbnail($scale, $win, $win->title(), $id_str, sub ($obj, $e) {
                     return unless $e->button == 1;
                     $cb->();
                 });
@@ -200,15 +235,27 @@ sub expose {
 # Inverse approach is used in order to simplify Expose deletion / re-implementation
 BEGIN {
     # Insert some pixbuf-specific methods 
-    sub X11::korgwm::Window::_update_pixbuf($self) {
-        return $self->{pixbuf} = Gtk3::GdkPixbuf::Pixbuf->new_from_xpm_data(['1 1 1 1', 'a c #000000', 'a'])
+    sub X11::korgwm::Window::_get_pixbuf($self) {
+        # If the window was not mapped, draw it in black
+        return Gtk3::GdkPixbuf::Pixbuf->new_from_xpm_data(['1 1 1 1', 'a c #262729', 'a'])
             unless $self->{real_w} and $self->{real_h};
-        my $win = Gtk3::Gdk::X11Window->foreign_new_for_display($display, $self->{id});
-        $self->{pixbuf} = Gtk3::Gdk::pixbuf_get_from_window($win, 0, 0, @{ $self }{qw( real_w real_h )});
-    }
 
-    # Register hide hook
-    push @X11::korgwm::Window::hooks_hide, sub($self) { $self->_update_pixbuf(); };
+        # This routine gets RGBA 24TT image but Gtk3 cat convert it only to 8-bit Pixbuf :(
+        my $pixmap = $X->generate_id();
+        $X->composite_name_window_pixmap($self->{id}, $pixmap);
+        my $image = $X->get_image(IMAGE_FORMAT_Z_PIXMAP, $pixmap, 0, 0, $self->{real_w}, $self->{real_h}, -1);
+        $image = $X->get_image_data_rgba($image->{sequence});
+
+        return Gtk3::GdkPixbuf::Pixbuf->new_from_bytes(
+            Glib::Bytes->new($image->{data}),   # data
+            "rgb",                              # colorspace = RGB
+            1,                                  # has alpha
+            8,                                  # ASSertion bits_per_sample == 8 are you kidding?
+            $self->{real_w},                    # width
+            $self->{real_h},                    # height
+            $self->{real_w} * 4,                # distance in bytes between rows aka rowstride
+        );
+    }
 }
 
 sub init {
